import fs from 'node:fs';
import nodePath from 'node:path';
import { finished } from 'stream/promises';

import { writeFileAndCreateFolders } from './helpers.ts';

import type { SongMeta, ServerCreds, Playlist } from './services/jellyfin.ts';
import { getSongFileBuffer } from './services/jellyfin.ts';

export function getLocalPath(serverPath: string) {
  const remoteRoot = process.env.REMOTE_ROOT_PATH as string;

  return nodePath.dirname(nodePath.normalize(serverPath).slice(remoteRoot.length));
}

export function getSongLocalFileName({ AlbumArtist, Artists, Name, Path }: SongMeta) {
  if (!Name) {
    return nodePath.basename(Path);
  }
  const ext = Path.split('.').pop();
  let artist = AlbumArtist || Artists?.[0];
  if (!artist) {
    return `${Name}.${ext}`;
  }
  return `${artist} - ${Name}.${ext}`;
}

export function downloadsFromPlaylists(playlists: Playlist[]) {
  const dedupedSongs: SongMeta[] = Object.values(
    playlists.reduce(
      (songs, pl) => {
        pl.songList?.forEach(song => {
          songs[song.Id] = song;
        });
        return songs;
      },
      {} as Record<string, SongMeta>,
    ),
  );

  return dedupedSongs.map(s => ({
    id: s.Id,
    downloadPath: getLocalPath(s.Path),
    localFileName: getSongLocalFileName(s),
  }));
}

const QUEUE: (() => Promise<void>)[] = [];

let downloadCount = 0;
const DOWNLOAD_CONCURENCY_LIMIT = 10;
const DOWNLOADED: Record<string, boolean> = {};

export const downloadSongs = async (
  creds: ServerCreds,
  songs: ReturnType<typeof downloadsFromPlaylists>,
  localRootPath: string,
) => {
  const promises = songs.map(s => getSongDownloader(creds, s, localRootPath));
  QUEUE.push(...promises);
  processQueue();
  await Promise.all(promises);
};
//TODO:   Now we have the local path and file name.  We need to check if the song has already downloaded and of not, start the download and add it to the list of downloaded, otherwise return false
const getSongDownloader =
  (
    creds: ServerCreds,
    song: ReturnType<typeof downloadsFromPlaylists>[number],
    localRootPath: string,
  ) =>
  async () => {
    const filePath = nodePath.join(localRootPath, song.downloadPath, song.localFileName);
    if (DOWNLOADED[filePath]) {
      return;
    }
    DOWNLOADED[filePath] = true;
    try {
      const download = await getSongFileBuffer(creds, song.id);

      console.log('Downloading song: ', song.localFileName);

      await writeFileAndCreateFolders(filePath, download);
    } catch (err: any) {
      if (err?.code !== 'EEXIST') {
        throw err;
      }
      console.log('File already exists, skipping: ', filePath);
    }
  };

const processQueue = () => {
  if (downloadCount < DOWNLOAD_CONCURENCY_LIMIT && QUEUE.length) {
    downloadCount++;
    const next = QUEUE.shift();
    next &&
      next().then(() => {
        downloadCount--;
        processQueue();
      });
    processQueue();
  }
};
